﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Diagnostics;


namespace GraphSD
{

    public enum RSubGraphHeuristicType { 
        MaxSupply,
        MinNeighbors,
        BalanceSupNeigh,
        MinSupply 
    }

    public enum RNodeHeuristicType
    {
        MaxDemand, 
        MaxNeighbors, 
        BalanceDemNeigh,
        MinDemand 
    }
    public enum RCorrectionType
    {
        None,
        Basic,
        Complete
    }
    
    
    public class GraphPartitioningSolvedProblemSD {

        public RGraphSD              mGraph;
        public RPartitioningGraphSD  mPartitioning;

        public GraphPartitioningSolvedProblemSD() { 
        
        }
    }
    public class GraphPartitioningProblemSD
    {
        RGraphSD               mOriginalGraph;
        RGraphSD               mCurrentGraph;
        RPartitioningGraphSD   mPartitioning;
        VisualiseGraphProblem  mVisualiser;
        Random                 mGenerator;
        int                    mCoveredDemand;
        int                    mTotalSupply;
        String                 mFileName;
        RSubGraphHeuristicType mHeuristicSubgraph;
        RNodeHeuristicType     mHeuristicNode;
        bool                   mDeepCorrection;
        RCorrectionType        mUseCorrection;
        bool                   mUseMultiHeuristic;
        long                   mCalcTime;
        

        RTabu                  mTabu;

        public long CalcTime {

            get { return mCalcTime; }
            set { mCalcTime = value; }

        }
        public RCorrectionType UseCorrection
        {

            get { return mUseCorrection; }
            set { mUseCorrection = value; }
        }

        public bool UseMultiHeuristic
        {

            get { return mUseMultiHeuristic; }
            set { mUseMultiHeuristic = value; }
        }

        public bool DeepCorrection {

            get { return mDeepCorrection; }
            set {  mDeepCorrection = value; }
        
        }

        public RSubGraphHeuristicType HeuristicSubgraph {

            get { return mHeuristicSubgraph; }
            set { mHeuristicSubgraph = value; } 
        }

        public RNodeHeuristicType HeuristicNode
        {

            get { return mHeuristicNode; }
            set { mHeuristicNode = value; }
        }

        public String FileName {
            get { return mFileName; }
            set { mFileName = value; }
        }

        RGraphSD GenerateGraph(int nNumberOfSupply, int nNumberOfDemand, int nNumberOfEdges) {

            RGraphSD            tGraph = new RGraphSD(nNumberOfSupply, nNumberOfDemand);
            int                 tMax = 8;

            RPartitioningGraphSD tPartitioning = new RPartitioningGraphSD();

            tGraph.GenerateRandomEdges(nNumberOfEdges);
            tGraph.GenrateRandomNegativeWeights(1, tMax);


            tPartitioning.SetGraph(tGraph);
            tPartitioning.SetOriginalGraph(tGraph);

            tPartitioning.GenerateCorrectedOptimalPartitioning(1,8, (float) 0.9);

            return tGraph;
        
        }

        public int CoveredDemand{
            get { return mCoveredDemand; }
        }

        public int TotalSupply{
            get { return mTotalSupply; }
        }
        public GraphPartitioningProblemSD() {

   //         mHeuristicSubgraph = RSubGraphHeuristicType.MaxSupply;
         mHeuristicSubgraph = RSubGraphHeuristicType.MinNeighbors;
         mHeuristicNode = RNodeHeuristicType.MaxDemand;
         mDeepCorrection = true;
         mUseMultiHeuristic = false;
  //        mHeuristicSubgraph = RSubGraphHeuristicType.BalanceSupNeigh;

            mFileName = "test_5_15_23.gpr";
            mVisualiser = new VisualiseGraphProblem();
            mPartitioning = new RPartitioningGraphSD();
            mOriginalGraph = new RGraphSD(5,26);

            mOriginalGraph.LoadGraph(mFileName);

            mCurrentGraph = new RGraphSD();

            mGenerator = new Random(1);
           // mOriginalGraph.SetGenerator(mGenerator);
           // mOriginalGraph.GenerateRandomEdges(60);
            mCurrentGraph.SetGenerator(mGenerator);
            

         //   mOriginalGraph.GenerateRandomWeights();
         //   mCurrentGraph.Copy(mOriginalGraph);

           
            InitializeVisualiser();
            mTabu = new RTabu();
            //mCoveredDemand = 0;
            //mTotalSupply   = 0;

        
        }

        public void InitializeVisualiser() {

            mVisualiser.SetGraph(mOriginalGraph);
            mVisualiser.GeneratePositions();
            mVisualiser.SetPartitioning(mPartitioning);
            
        } 

        public VisualiseGraphProblem Visualiser {
            get { return mVisualiser; }
            set { mVisualiser = value; }
        }

        public void StartCalculation(){
            RGraphSD Temp;


            mOriginalGraph.LoadGraph(mFileName);
            mCurrentGraph = new RGraphSD();
            mCurrentGraph.Copy(mOriginalGraph);

           // mPartitioning.Clear();
         
            mPartitioning.SetGraph(mCurrentGraph);
            mPartitioning.SetOriginalGraph(mOriginalGraph);

            mPartitioning.SetGenerator(mGenerator);
            
            InitializeVisualiser();
            mPartitioning.Initialize();


            mCoveredDemand = 0;
            mTotalSupply = 0;

            


            
     //       mCurrentGraph.RemoveAllSupplyConnections();
        }


        public RSubGraphSD GetHeuristicSubgraphs() {
            RSubGraphSD result = null;


            switch (mHeuristicSubgraph){
            
                case RSubGraphHeuristicType.MaxSupply:
                    return mPartitioning.GetSubgraphWithMaxSupply();
                case RSubGraphHeuristicType.MinSupply:
                    return mPartitioning.GetSubgraphWithMinSupply();

                case RSubGraphHeuristicType.MinNeighbors:
                         return mPartitioning.GetSubgraphWithMinNieghbors();
               
                case RSubGraphHeuristicType.BalanceSupNeigh:
                         return mPartitioning.GetSubgraphWithMaxBalanceNeighSup();


            }
            return result;
        }

        public int GetHeuristicNode(RSubGraphSD mSubgraph)
        {
            int result = -1;


            switch (mHeuristicNode)
            {

                case RNodeHeuristicType.MaxDemand:
                    return mSubgraph.GetHighestDemand();
                case RNodeHeuristicType.MinDemand:
                    return mSubgraph.GetLowestDemand();

                case RNodeHeuristicType.MaxNeighbors:
                    return mSubgraph.GetMaxNeighbors(mPartitioning.GetSubGraphNodes());

                case RNodeHeuristicType.BalanceDemNeigh:
                    return mSubgraph.GetMaxBalanceDemNeigh(mPartitioning.GetSubGraphNodes());


            }
            return result;
        }
     

        public void GreedyCalcPartioning(int NumberOfSteps = -1)
        {
            RSubGraphSD cSubGraph;
            int cDemmandNode;
            int StepsLeft = NumberOfSteps;
            mTabu.Empty();
            while (mPartitioning.AvailableSupply > 0) {

//                cSubGraph = mPartitioning.GetSubgraphWithMaxSupply();
             //   cSubGraph = mPartitioning.GetSubgraphWithMaxSupply();
                cSubGraph = GetHeuristicSubgraphs();
                

                if (cSubGraph == null)
                    break;
//                cDemmandNode = cSubGraph.GetHighestDemand();
                cDemmandNode = GetHeuristicNode(cSubGraph);
//                    cSubGraph.GetHighestDemand();


                if (cDemmandNode != -1)
                {
                    mPartitioning.AddNodeToSubGraph(cDemmandNode, cSubGraph.SupplyNodeId);
                    mCurrentGraph.RemoveConnections(cDemmandNode);
                    mPartitioning.RemoveNeighbor(cDemmandNode, mPartitioning.GetPartitionIndexForNode(cSubGraph.SupplyNodeId));
                    StepsLeft--;

                }
                else {

                    mPartitioning.CloseSubGraph(mPartitioning.GetPartitionIndexForNode(cSubGraph.SupplyNodeId));
                }

                if (StepsLeft == 0)
                    break;
            }

            mCoveredDemand = mPartitioning.GetCoveredDemmand();
            mTotalSupply = mOriginalGraph.GetTotalSupply();
        }

        public static GraphPartitioningSolvedProblemSD GenerateSolvedGraph(int nNumSupply, int nNumDemmand, int nNumEdges, int nMinWeight, int nMaxWeight, float nCovered, int nRandomSeed, bool tree)
        {
            RGraphSD Result = new RGraphSD();


            RGraphSD tOriginalGraph = new RGraphSD(nNumSupply, nNumDemmand);
            RGraphSD tCurrentGraph = new RGraphSD();
            RPartitioningGraphSD tPartitioning = new RPartitioningGraphSD();
            RGraphSD Temp = new RGraphSD();

            Random tGenerator = new Random(nRandomSeed);

            tOriginalGraph.SetGenerator(tGenerator);
            if (tree)
                    tOriginalGraph.GenerateRandomTreeEdges();
            else
                 tOriginalGraph.GenerateRandomEdges(nNumEdges);
        
            tCurrentGraph = new RGraphSD();
            tCurrentGraph.Copy(tOriginalGraph);
            tPartitioning.SetGraph(tCurrentGraph);
            tPartitioning.SetOriginalGraph(tOriginalGraph);

            tPartitioning.SetGenerator(tGenerator);
            Temp = tPartitioning.GenerateCorrectedOptimalPartitioning(nMinWeight, nMaxWeight, nCovered);
            if (Temp == null)
                return null;

            tOriginalGraph.CopyWeights(Temp);

            GraphPartitioningSolvedProblemSD Solved = new GraphPartitioningSolvedProblemSD();

            Solved.mGraph        = tOriginalGraph;
            Solved.mPartitioning = tPartitioning;


            return Solved;

        }


        public static String GenerateFileNameProblems(int nNumSupply, int nNumDemmand, int nNumEdges, int nRandomSeed){
        
                    String Temp;

                    Temp = "test_" + nNumSupply + "_" + nNumDemmand + "_" + nRandomSeed + ".gpr";

                    return Temp;
            }
          public static String GenerateFileNameSolution(int nNumSupply, int nNumDemmand, int nNumEdges,  int nRandomSeed){
        
                    String Temp;

                    Temp = "sol_" + nNumSupply + "_" + nNumDemmand + "_" + nRandomSeed + ".sol";

                    return Temp;
        
        }

          public static void GenerateTestsSet(int nNumSupply, int nNumDemmand, int nNumEdges, int nMinWeight, int nMaxWeight, float nCovered, bool tree)
          {


            GraphPartitioningSolvedProblemSD tSolved;
            String          tString;
            int counter = 0;
            int NonEmpty;
            int index;
            int Seed;
            Seed = 0;

            double tolerance;
                while(counter <40){

                    tSolved = GraphPartitioningProblemSD.GenerateSolvedGraph(nNumSupply, nNumDemmand, nNumEdges, nMinWeight, nMaxWeight, (float)0.9, Seed, tree);

                    if (tSolved != null) { 
    
                        tString = GenerateFileNameProblems(nNumSupply, nNumDemmand, nNumEdges, counter);
                        if (tree)
                           tString = ".\\Tree\\" + tString;
                        else
                            tString = ".\\General\\" + tString;

                        tSolved.mGraph.SaveGraph(tString);

                        tString = GenerateFileNameSolution(nNumSupply, nNumDemmand, nNumEdges, counter);

                        if (tree)
                            tString = ".\\Tree\\" + tString;
                        else
                            tString = ".\\General\\" + tString;

                        tSolved.mPartitioning.Save(tString);
                        counter++;
                    }
                    Seed++;
                }
                    
        
        }

        public int SolveProblemMultiHeuristic()
        {

            int BestResult = int.MaxValue;
            int CurrentResult = int.MaxValue;

                  
                for(int i =0; i<4; i++)
                    for (int j = 0; j < 3; j++) {

                        mHeuristicNode = (RNodeHeuristicType)i;
                        mHeuristicSubgraph = (RSubGraphHeuristicType)j;

                        CurrentResult = SolveProblem();
                        if (CurrentResult < BestResult)
                                  BestResult = CurrentResult;
                    }

                return BestResult;
        }

        public int SolveProblem() {

            int Result;
            int Value, Value1;
            int Sum = 0;
            long trun;

            DateTime StartTime = DateTime.Now;

            //Code
            Stopwatch stopWatch = new Stopwatch();

            StartCalculation();
            stopWatch.Restart();

            stopWatch.Start();
 
            GreedyCalcPartioning(-500);

            Value = mPartitioning.GetCoveredDemmand();


            if (mUseCorrection == RCorrectionType.Basic)
                              GreedyCalcCorrectionBasic(100);
            if (mUseCorrection == RCorrectionType.Complete)
                        GreedyCalcCorrectionComplete(100);
            
            Value1 = mPartitioning.GetCoveredDemmand();
//          if (Value != Value1)
//                  Sum += Value1 - Value;

            Result = mPartitioning.GetCoveredDemmand();


            stopWatch.Stop();
            trun = stopWatch.ElapsedMilliseconds;
            mCalcTime += stopWatch.ElapsedMilliseconds;
            
            return Result;
        
        
        }

        public void  SolveProblemSet(int nNumSupply, int nNumDemmand, out List<int> Results, out List<int> Opt, string Folder) {


            String tString;
            Results = new List<int>();
            Opt = new List<int>();
            int nNumEdges=0;
            int Sum = 0;
            int Solution;

            int Value, Value1;
            mCalcTime = 0;

            for (int i = 0; i < 40; i++) {

                mFileName =Folder + GenerateFileNameProblems(nNumSupply, nNumDemmand, nNumEdges, i);

                if(mUseMultiHeuristic)
                    Solution = SolveProblemMultiHeuristic();
                else
                    Solution = SolveProblem();
                Results.Add(Solution);
//                Results.Add(mPartitioning.GetCoveredDemmand());

                Opt.Add(TotalSupply);

                /*
                StartCalculation();
                GreedyCalcPartioning(-500);

                Value = mPartitioning.GetCoveredDemmand();
                if (mUseCorrection)
                  //  GreedyCalcCorrectionBasic(100);
                    GreedyCalcCorrectionAdvance(100);

                Value1 = mPartitioning.GetCoveredDemmand();
                if (Value != Value1)
                    Sum +=  Value1 - Value;

                Results.Add(mPartitioning.GetCoveredDemmand());
                Opt.Add(TotalSupply);
                */


            }
                
        }

        public  bool GreedyCalcCorrectionBasic(int Steps) {
            int Const = 0;
            int PrevDemand = 0;

            for (int i = 0; i < Steps; i++) {

                if (!GreedyCalcCorrection()) {
                    if (!GreedyCalcCorrection(true))
                        return false;
                }
                if (PrevDemand == mCoveredDemand)
                    Const++;
                else{
                    Const = 0;
                    PrevDemand = mCoveredDemand;
                }
                if (Const == 3)
                    break;
            }
            return true;
        }

        public bool GreedyCalcCorrectionComplete(int Steps)
        {

            bool ChangeBasic = false;
            bool ChangeLoc = false;

            bool ChangeAdvance = false;
            int  Counter = 0;
            ;
            int  MaxAdvance = 2;
            int  BestCover = mPartitioning.GetCoveredDemmand();
            int  CurrentCover;
            int  ConstCounter=0;
            int  MaxConst = 3;
            int  PrevCover;
            int CounterLoop2 = 0;
            int i;

            ConstCounter = 0;

            for ( i = 0; i < Steps; i++)
            {
                ChangeBasic = GreedyCalcCorrectionBasic(Steps);
                ChangeAdvance = false;
                ChangeLoc = true;
                Counter = 0;
                PrevCover = mPartitioning.GetCoveredDemmand();

                while (ChangeLoc && MaxAdvance > Counter)
                {
                    ChangeLoc = GreedyCalcCorrectionAdvance(Steps);
                    ChangeAdvance = ChangeAdvance || ChangeLoc;
                    // if (ChangeAdvance)
                    CurrentCover = mPartitioning.GetCoveredDemmand();
                    if (PrevCover == CurrentCover)
                        Counter++;
                    else
                        Counter=0;

                    PrevCover = CurrentCover;
                    CounterLoop2++;
                }

                CurrentCover = mPartitioning.GetCoveredDemmand();

                if (CurrentCover < BestCover)
                {
                    BestCover = CurrentCover;
                    ConstCounter = 0;
                }
                else {

                    ConstCounter++;
                }

                if (!ChangeAdvance && !ChangeBasic)
                    break;

                if (ConstCounter >= MaxConst)
                    break;
            }

            mCoveredDemand = mPartitioning.GetCoveredDemmand();
            mTotalSupply = mOriginalGraph.GetTotalSupply();

            return true;
        }


        public bool GreedyCalcCorrectionAdvance(int Steps)
        {

            List<int> NeighboringLeafs;
            List<int> Except = new List<int>();
            List<int> ExceptNodes = new List<int>();

            List<int> MaxMissingDemandSubgraph;
            List<int> BestCoverNode;
            bool More = true;
            bool NoChange = true;
            bool HadEffect = false;
            int  CurrentTest = 0;
         //   for (int i = 0; i < Steps; i++)
            //{

            mPartitioning.GetAllNeighboringSubGraphs();

            CurrentTest = 0;
            while (More) { 

//                mPartitioning.GetAllNeighboringSubGraphs();
                MaxMissingDemandSubgraph = mPartitioning.GetSubgraphWithMaxMissingDemand(Except);
                if (MaxMissingDemandSubgraph.Count <= CurrentTest)
                    break;

                
                NoChange = true;
                while (NoChange)
                {
                    // Just ajusting the heuristic used
                    if(!DeepCorrection)
                                NoChange = false;

//                    mPartitioning.GetAllNeighboringSubGraphs();
                    ExceptNodes = mTabu.GetForbidden(MaxMissingDemandSubgraph[CurrentTest]);
                    NeighboringLeafs = mPartitioning.GetAllNeighboringLeafs(MaxMissingDemandSubgraph[CurrentTest]);
                    BestCoverNode = mPartitioning.GetBestCoverLeafNode(MaxMissingDemandSubgraph[CurrentTest], ExceptNodes);

                    if (BestCoverNode.Count > 0)
                    {
                        if (MaxMissingDemandSubgraph[CurrentTest] != -1)
                            mTabu.Add(BestCoverNode[0], mPartitioning.GetPartitionIndexForSupplyNode(BestCoverNode[0]), MaxMissingDemandSubgraph[CurrentTest]);
                        mPartitioning.RemoveNodeFromSubgraph(BestCoverNode[0]);
                        UpdateTreeForRemovedNode(BestCoverNode[0]);
                        ChangeNodeInSubgraph(BestCoverNode[0], -1, MaxMissingDemandSubgraph[CurrentTest]);
                        More = false;
                        HadEffect = true;
                        mPartitioning.UpdateConnectedNeighboringSubGraphs(BestCoverNode[0]);
                    }
                    else
                    {
                        NoChange = false;
                        Except.Add(MaxMissingDemandSubgraph[CurrentTest]);
                        CurrentTest++;
                    //    if (!HadEffect)
                                 More = true;
                    }
                }
            }

           //     for (int j = 0; j < mPartitioning.NumberOfSubGraphs; j++ )
           //         
                
                /*
                if (!GreedyCalcCorrection())
                {
                    if (!GreedyCalcCorrection(true))
                        return false;
                }
                 */
           // }

                mCoveredDemand = mPartitioning.GetCoveredDemmand();
                mTotalSupply = mOriginalGraph.GetTotalSupply();

                return HadEffect;
        }

        public bool GreedyCalcCorrection(bool OnlySwitch = false) {

            RSubGraphNodeArray tSubGraphNodes;
            List<int>       tNonLoc = new List<int>();
            List<int>       tConSubgraphs;
            List<int>       tLeafs;
            
            List<int>[]       aLeafs = new List<int>[mPartitioning.NumberOfSubGraphs];
            bool[]            lCalculated = new bool[mPartitioning.NumberOfSubGraphs];
            int             Max;
            int             Index;
            int             MissingDemand;
            int             NonLocDemand;

            int             BestChange = -1; ;
            int             IdBest   = -1;
            int             LeafDemand;
            int             change;
            RSubGraphSD     tSubGraph;
            List<int>       tConnectedCheck;
            bool            IsRepeat;

            for (int i = 0; i < mPartitioning.NumberOfSubGraphs; i++)
                lCalculated[i] = false;
                tSubGraphNodes = mPartitioning.GetSubGraphNodes();

            tNonLoc = tSubGraphNodes.GetNotInSubgraph();

            foreach (int nl in tNonLoc) {

                tConSubgraphs = tSubGraphNodes.GetConnectedSubgraphs(nl, mOriginalGraph);

                NonLocDemand = mOriginalGraph.GetWeight(nl);

                foreach (int sl in tConSubgraphs)
                {
                    if (!lCalculated[sl])
                    {
                        aLeafs[sl] = tSubGraphNodes.GetLeafs(sl, mPartitioning);
                        lCalculated[sl] = true;
                    }
                    tLeafs = aLeafs[sl];// tSubGraphNodes.GetLeafs(sl, mPartitioning);
                    tSubGraph   = mPartitioning.GetSubGraph(sl);


                    MissingDemand = mPartitioning.GetSubGraph(sl).GetCoveredDemand();
                    MissingDemand += mOriginalGraph.GetWeight(tSubGraph.SupplyNodeId);

                    if ( (MissingDemand >= -NonLocDemand)){

                        ChangeNodeInSubgraph(nl, -1, sl);

                        mCoveredDemand = mPartitioning.GetCoveredDemmand();
                        mTotalSupply = mOriginalGraph.GetTotalSupply();

                        return true;
                    }


                    foreach (int l in tLeafs) {

                        LeafDemand = mOriginalGraph.GetWeight(l);

                        change = -NonLocDemand + LeafDemand;
                        IsRepeat = false;
                        if (OnlySwitch && (change == 0))
                        {

                            if(mTabu.IsLoop(5))
                                    IsRepeat = mTabu.InLastMoves(l,5);
                        }
                        if ((change <= MissingDemand) &&(change >BestChange)
                            && (((change > 0) && !OnlySwitch) || (OnlySwitch && (change == 0)&&!IsRepeat))
                            )
                        {
                            tConnectedCheck = tSubGraph.GetConnectedNodes(nl, mOriginalGraph);

                            if ((tConnectedCheck.Count > 1) || (tConnectedCheck[0] != l))
                            {
                                IdBest = l;
                                BestChange = change;
                            }
                        } 
                    }

//////////////        Check if  Found Improvement

                    if (IdBest != -1) {

                        if (OnlySwitch)
                        {
                            mTabu.Add(nl, mPartitioning.GetPartitionIndexForNode(nl), sl);
                            mTabu.Add(IdBest, sl,mPartitioning.GetPartitionIndexForNode(IdBest));

                        }
                        ChangeNodeInSubgraph(nl, IdBest, sl);
                         mCoveredDemand = mPartitioning.GetCoveredDemmand();
                        mTotalSupply = mOriginalGraph.GetTotalSupply();

                        return true;
                    }
                      
                }
            }
            mCoveredDemand = mPartitioning.GetCoveredDemmand();
            mTotalSupply = mOriginalGraph.GetTotalSupply();

            return false;
        }


        public void UpdateTreeForRemovedNode(int RemoveNode) { 
            List<int> tConnected;
            RSubGraphSD tSubgraph;
            RSubGraphNodeArray tSubNodes;
            int       MaxTree = 0;
            int       IdMax   = -1;

            if (RemoveNode != -1)
            {
                
               // tSubgraph = mPartitioning.GetSubGraph(mPartitioning.GetPartitionIndexForNode(RemoveNode));
                tSubgraph = mPartitioning.GetSubGraph(mPartitioning.GetPartitionIndexForSupplyNode(RemoveNode));
                if (tSubgraph != null)
                {
                    tConnected = tSubgraph.GetConnectedNodes(RemoveNode, mOriginalGraph);
                }
                else
                    tConnected = new List<int>();
                tSubNodes = mPartitioning.GetSubGraphNodes();
                tSubNodes[RemoveNode].SubGraphId = -1;

                foreach (int con in tConnected)
                {

                    if (tSubNodes[con].TreeConnect >= MaxTree)
                    {

                        if ((tSubNodes[con].TreeConnect > MaxTree) ||
                            ((tSubNodes[con].TreeConnect == MaxTree) && (con < IdMax)))
                        {
                            MaxTree = tSubNodes[con].TreeConnect;
                            IdMax = con;
                        }
                    }
                }




                if (IdMax != -1)
                {
                    if (mOriginalGraph.GetWeight(IdMax) < 0)
                    {
                        tSubNodes[IdMax].TreeConnect--;
                        //      tSubNodes[IdMax].SubGraphId = -1; 
                    }
                }
                else
                {
                    ;
                }
            }
        }
        //
        //          Currently does not fix current graph and the neighbors in subgraph
        //          it is only used for the correction
        //
        //
        public void ChangeNodeInSubgraph(int AddNode, int RemoveNode, int IdSubgraph) {

            RSubGraphSD tSubgraph;
            List<int> tConnected;
            int       MaxTree = 0;
            int       IdMax   = -1;
            RSubGraphNodeArray tSubNodes;

            tSubgraph  = mPartitioning.GetSubGraph(IdSubgraph);


            
//// Remove Node            
            tSubgraph.RemoveNode(RemoveNode);

             UpdateTreeForRemovedNode(RemoveNode);
            /*
            
            if (RemoveNode != -1) {

            tConnected = tSubgraph.GetConnectedNodes(RemoveNode, mOriginalGraph);

            tSubNodes = mPartitioning.GetSubGraphNodes();
            tSubNodes[RemoveNode].SubGraphId = -1;

            foreach (int con in tConnected) {

                if (tSubNodes[con].TreeConnect >= MaxTree) { 
                
                    if((tSubNodes[con].TreeConnect > MaxTree)||
                      ((tSubNodes[con].TreeConnect == MaxTree)&&(con < IdMax)))
                    {
                            MaxTree = tSubNodes[con].TreeConnect;
                            IdMax = con;
                    }
                }
            }
            
            if (IdMax != -1)
            {
                if (mOriginalGraph.GetWeight(IdMax) < 0) {
                    tSubNodes[IdMax].TreeConnect--;
              //      tSubNodes[IdMax].SubGraphId = -1; 
                }
            }
            else {
                ;  
            }
              
            }
 */
///// Add Node
            mPartitioning.AddNodeToSubGraph(AddNode, tSubgraph.SupplyNodeId);
            
        }

        
    }

    
   
}
