﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GraphSD
{
    public class RPartitioningGraphSD
    {

        RGraphSD             mGraph;
        RGraphSD             mOriginalGraph;
        
        int                  mNumberOfSubGraphs;
        RSubGraphSD[]        mPartitioning;
        int                  mAvailableSupply;
        Random               mGenerator;
        RSubGraphNodeArray   mSubgraphNodes;
        int                  rMaxInt;
 


        public int GetTotalCoveredSupply() {
            int         Result = 0;
            List<int> Temp;

            for (int i = 0; i < mNumberOfSubGraphs; i++) {

                Temp = mPartitioning[i].GetInside();

                foreach (int elem in Temp) {

                    if (mGraph.GetWeight(elem) < 0)
                        Result -= mGraph.GetWeight(elem);
                }

            }
            return Result;
        }


        public int GetCoveredDemmand()
        {


            int Result = 0;
            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                Result += mPartitioning[i].GetCoveredDemand();

            }

            return Result;
        }

        public int NumberOfSubGraphs
        {

            get { return mNumberOfSubGraphs; }
        }
        public void RemoveAllSupplyNeighbors()
        {
            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                RemoveNeighbor(mPartitioning[i].SupplyNodeId, -1);
            }
        }

        
        public void SetGenerator(Random nGenerator) { mGenerator = nGenerator; }


        public RSubGraphNodeArray GetSubGraphNodes()
        {

            return mSubgraphNodes;
        }

        public bool CanBeNewSupply(int index, List<int> S, List<int>D, out int Dem) {


            List<int> Con;

            Dem = -1;
            if (S.Contains(index) || D.Contains(index))
                return false;

            Con = mGraph.GetConnections(index);
            
            foreach (int n in Con) {

                if (!(S.Contains(n) || D.Contains(n)))
                {
                    Dem = n;
                    return true;
                }
            }

            return false;
        
        }
        public RGraphSD GenerateCorrectedOptimalPartitioning(int nMinWeight,  int MaxWeight, float Covered)
        {


            mPartitioning = new RSubGraphSD[mGraph.NumberSupply];
            mNumberOfSubGraphs = 0;
            int index;
            bool isNew;
            List<int> NN;
            bool FreeNeighbors;
            bool NotInside;
            mGraph.SetGenerator(mGenerator);
            mGraph.GenrateRandomNegativeWeights(nMinWeight, MaxWeight);

            List<int> S = new List<int>();
            List<int> D = new List<int>();
            int dem;
            
            int found = 0;
            int sc = 0;
            while (found < mGraph.NumberSupply) {

                index = mGenerator.Next(mGraph.NumberOfVertices);

                if(CanBeNewSupply(index, S,D,out dem)){
                    S.Add(index);
                    D.Add(dem);
                    found++;
                }

                sc++;
                if (sc > 100000)
                    return null;
            }

            mNumberOfSubGraphs = 0;

            for (int i = 0; i < mGraph.NumberSupply; i++) {

                mPartitioning[mNumberOfSubGraphs] = new RSubGraphSD();
                mPartitioning[mNumberOfSubGraphs].SetGraph(mGraph);
                mPartitioning[mNumberOfSubGraphs].SetSupplyNode(S[i]);

                mPartitioning[mNumberOfSubGraphs].InitializeSupplyId(false);

                RemoveNeighbor(S[i], mNumberOfSubGraphs);
                mGraph.RemoveConnections(S[i]);


                mNumberOfSubGraphs++;
            }
            /*

                while (mNumberOfSubGraphs < mGraph.NumberSupply)
                {


                    index = mGenerator.Next(mGraph.NumberOfVertices);

                    isNew = true;
                    FreeNeighbors = true;

                    for (int j = 0; j < mNumberOfSubGraphs; j++)
                    {
                        if (mPartitioning[j].SupplyNodeId == index)
                        {
                            isNew = false;
                            break;
                        }


                    }

                    //check if blocked

                    FreeNeighbors = false;
                    NN = mGraph.GetConnections(index);
                    if (isNew)
                    {
                        foreach (int n in NN)
                        {
                            NotInside = true;
                            for (int k = 0; k < mNumberOfSubGraphs; k++)
                                if (n == mPartitioning[k].SupplyNodeId)
                                {
                                    NotInside = false;
                                    break;
                                }
                            FreeNeighbors = FreeNeighbors || NotInside;
                        }
                    }


                    if ((isNew) && (FreeNeighbors))
                    {
                        mPartitioning[mNumberOfSubGraphs] = new RSubGraphSD();
                        mPartitioning[mNumberOfSubGraphs].SetGraph(mGraph);
                        mPartitioning[mNumberOfSubGraphs].SetSupplyNode(index);

                        mPartitioning[mNumberOfSubGraphs].InitializeSupplyId(false);

                        RemoveNeighbor(index, mNumberOfSubGraphs);
                        mGraph.RemoveConnections(index);


                        mNumberOfSubGraphs++;
                    }

                }
            */
            //  for (int i = 0; i < mNumberOfSubGraphs; i++) {
            //
            //                RemoveNeighbor(mPartitioning[i].SupplyNodeId, -1);
            //            }
            RemoveAllSupplyNeighbors();
            int counter = mNumberOfSubGraphs;
            int randomSub;
            int RandomNeighbor;
            List<int> Neighbors;
            int AddingNode;

            int NoAdd = 0;


            for (int i = 0; i < mGraph.NumberSupply; i++)
            {
                mPartitioning[i].AddDemand(D[i], false);
                for (int j = 0; j < mNumberOfSubGraphs; j++)
                {
                    RemoveNeighbor(D[i], j);

                }

                mGraph.RemoveConnections(D[i]);

                counter++;
            }

            

            while (counter < mGraph.NumberOfVertices * Covered)
            {
                randomSub = mGenerator.Next(mNumberOfSubGraphs);
                 Neighbors = mPartitioning[randomSub].GetNeighbors();
                if (Neighbors.Count() > 0)
                {
                    RandomNeighbor = mGenerator.Next(Neighbors.Count());

                    AddingNode = Neighbors[RandomNeighbor];

                    mPartitioning[randomSub].AddDemand(AddingNode, false);
                    for (int j = 0; j < mNumberOfSubGraphs; j++)
                    {
                        RemoveNeighbor(AddingNode, j);
                     
                    }

                    mGraph.RemoveConnections(AddingNode);

                    counter++;
                    NoAdd = 0;
                }
                else {

                    if (NoAdd > 1000)
                    {
                        if (mGraph.HasEdge() )
                            NoAdd = 0;
                        else
                            break;
                    }
                    NoAdd++;
                }
                

            }

            int Total;
            List<int> Inside;
            int nSupply;
            int tWeight;

            int Sum;

            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {


//                Total = mPartitioning[i].AvailableSupply;


                Total = 0;

                Inside = mPartitioning[i].GetInside();
                foreach (int elem in Inside)
                {
                    Total += mGraph.GetWeight(elem);
                }


                
                nSupply = Inside[mGenerator.Next(Inside.Count)];
                tWeight = mGraph.GetWeight(nSupply);
                Total -= tWeight;
                mPartitioning[i].SetSupplyNode(nSupply);
                mGraph.setWeight(nSupply, -Total);
                mPartitioning[i].AvailableSupply = -Total;


                Sum = 0;

                foreach (int elem in Inside)
                {
                    Sum += mGraph.GetWeight(elem);
                }


            }

            return mGraph;


        }
        public void RemoveNeighbor(int NodeId, int ExceptPartitionId)
        {
            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                if (i != ExceptPartitionId)
                    mPartitioning[i].RemoveNeighbor(NodeId);
            }

        }


        public RGraphSD GetGraph() { return mGraph; }
        public void CloseSubGraph(int Index)
        {

            mAvailableSupply -= mPartitioning[Index].AvailableSupply;
            mPartitioning[Index].AvailableSupply = 0;
            mPartitioning[Index].GetNeighbors().Clear();

        }
        public List<int> GetSubgraphWithMaxMissingDemand(List<int> Except) {

            int MaxMissing = int.MinValue;
            int MaxIndex = -1;
            int Missing;
            List<int> Result = new List<int>();

            for (int i = 0; i < mNumberOfSubGraphs; i++) {
                if (!Except.Contains(i))
                {
                    Missing = mGraph.GetWeight(mPartitioning[i].SupplyNodeId) + mPartitioning[i].GetCoveredDemand();

                    if (Missing == MaxMissing)
                    {
                        Result.Add(i);
                    }

                    if (Missing > MaxMissing)
                    {
                        Result.Clear();
                        Result.Add(i);
                        MaxMissing = Missing;
                    }
                }
            }

            return Result;
        }
        public void RemoveNodeFromSubgraph(int NodeId) {
            int SubgraphID;
            SubgraphID = mSubgraphNodes[NodeId].SubGraphId;
            if (SubgraphID != -1)  
                mPartitioning[SubgraphID].RemoveNode(NodeId);

        }
        public int GetPartitionIndexForSupplyNode(int Index) {

            return mSubgraphNodes[Index].SubGraphId;
        }
        
        public int GetPartitionIndexForNode(int Index)
        {

            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {
                if (mPartitioning[i].SupplyNodeId == Index)
                    return i;
            }

            return -1;
        }
        public void AddNodeToSubGraph(int cDemmandNode, int SupplyNodeId)
        {
            int SubGraphId = GetPartitionIndexForNode(SupplyNodeId);

            mPartitioning[SubGraphId].AddDemand(cDemmandNode);
            mAvailableSupply += mGraph.GetWeight(cDemmandNode);

            UpdateSubgraphNodes(SubGraphId, cDemmandNode);
        }
        public void GetAllNeighboringSubGraphs() {

           

            List<int> tConnections;
            List<int> nSg;
            for (int i = 0; i < mGraph.NumberOfVertices; i++) {

                tConnections = mOriginalGraph.GetConnections(i);
                mSubgraphNodes[i].NeighboringSubgraphs.Clear();
                foreach (int con in tConnections) {

                   
                       nSg = mSubgraphNodes[i].NeighboringSubgraphs;
                         if(!nSg.Contains(mSubgraphNodes[con].SubGraphId) &&                              (mSubgraphNodes[con].SubGraphId != mSubgraphNodes[i].SubGraphId))
                                nSg.Add(mSubgraphNodes[con].SubGraphId);
                         
                         
                    
                }
            }
         }

        public void UpdateConnectedNeighboringSubGraphs(int NodeID) {

            List<int> PotentialChanged = mOriginalGraph.GetConnections(NodeID);
            List<int> tConnections;
            List<int> nSg;
            PotentialChanged.Add(NodeID);

            foreach (int ch in PotentialChanged)
            {

                tConnections = mOriginalGraph.GetConnections(ch);
                mSubgraphNodes[ch].NeighboringSubgraphs.Clear();
                foreach (int con in tConnections)
                {


                    nSg = mSubgraphNodes[ch].NeighboringSubgraphs;
                    if (!nSg.Contains(mSubgraphNodes[con].SubGraphId) && (mSubgraphNodes[con].SubGraphId != mSubgraphNodes[ch].SubGraphId))
                        nSg.Add(mSubgraphNodes[con].SubGraphId);



                }
            }

        
        }

        public List<int> GetAllNeighboringLeafs(int SubGraphIndex) {

            List<int> Result = new List<int>();

            for (int i = 0; i < mGraph.NumberOfVertices; i++) {

        //        if ((mSubgraphNodes[i].TreeConnect < 1) && mSubgraphNodes[i].NeighboringSubgraphs.Contains(SubGraphIndex)
        //            && (mGraph.GetWeight(i)<=0))
                    if ((mSubgraphNodes[i].TreeConnect < 1)
                        && (mGraph.GetWeight(i) <= 0)) { 

                        if( mSubgraphNodes[i].NeighboringSubgraphs.Contains(SubGraphIndex))
                                             Result.Add(i);
                    }
            }

           return Result;
        
        }

        public List<int> GetBestCoverLeafNode(int SubgraphIndex, List<int> Except) {

            List<int> Result = new List<int>();
            List<int> NeighboringLeafs;

            int       Improve;
            int       BestIndex;
            int BestImprove = rMaxInt;// = 10000;
       //     int BestImprove = int.MaxValue;
        
            bool      BestIsUnlocated;
            bool      IsUnlocated;

            int       AvailableSupply;
            
            NeighboringLeafs = GetAllNeighboringLeafs(SubgraphIndex);
            
            AvailableSupply = mGraph.GetWeight(mPartitioning[SubgraphIndex].SupplyNodeId) + mPartitioning[SubgraphIndex].GetCoveredDemand();

            BestIsUnlocated = false;
            foreach (int leaf in NeighboringLeafs) { 
            
                    Improve = AvailableSupply + mGraph.GetWeight(leaf);
                    IsUnlocated = (mSubgraphNodes[leaf].SubGraphId == -1);

                    if ((!BestIsUnlocated) || (IsUnlocated))
                    {


                        if ((Improve >= 0) && ((Improve <= BestImprove) || (IsUnlocated && (!BestIsUnlocated))))
                        {

                            // if they are the same just add to list
                            if ((BestIsUnlocated == IsUnlocated) && (Improve == BestImprove)) {
                                if(!Except.Contains(leaf))
                                Result.Add(leaf);
                            }else{
                                       //if the new leaf is unlocaed it always has an advantage if the previose was located
                                    if (((!BestIsUnlocated) && (IsUnlocated))
                                        // they are both located or non located just one has a better improvement
                                        ||((BestIsUnlocated == IsUnlocated) && (Improve < BestImprove)))
                                    {

                                        if (!Except.Contains(leaf))
                                        {
                                            Result.Clear();
                                            Result.Add(leaf);
                                            BestImprove = Improve;
                                            BestIsUnlocated = IsUnlocated;
                                            BestIndex = leaf;
                                        }
                                    }                                    
                             }

                            
                            
                                    
                         }

                  }
             }
            
 
            return Result;

        
        }
        public void UpdateSubgraphNodes(int SubGraphId, int cDemmandNode){
            List<int> tConnections = new List<int>();
            int       MaxConnections = -1;
            int       IdMaxConnections = -1;

            mSubgraphNodes[cDemmandNode].SubGraphId = SubGraphId;
            tConnections = mOriginalGraph.GetConnections(cDemmandNode);

            foreach(int con in tConnections){
                    
                   if(mSubgraphNodes[con].SubGraphId == SubGraphId){
                   
                        if(mGraph.GetWeight(con)>0){
                                IdMaxConnections = con;
                                MaxConnections = int.MaxValue;
                        
                        }
                        if((MaxConnections<mSubgraphNodes[con].TreeConnect)
                        || ((MaxConnections==mSubgraphNodes[con].TreeConnect) &&(con<IdMaxConnections)))
                       {

                            if (MaxConnections < mSubgraphNodes[con].TreeConnect) 
                            {
                                MaxConnections = mSubgraphNodes[con].TreeConnect;
                                IdMaxConnections = con;
                            }
                        }
                   }
            }


            if((MaxConnections< int.MaxValue)  && (IdMaxConnections !=  -1) )
                mSubgraphNodes[IdMaxConnections].TreeConnect++;
            
        }
        public int AvailableSupply
        {

            get { return mAvailableSupply; }
            set { mAvailableSupply = value; }

        }

        public void Initialize()
        {


            mPartitioning = new RSubGraphSD[mGraph.NumberSupply];
            mNumberOfSubGraphs = 0;
            mAvailableSupply = 0;

            for (int i = 0; i < mGraph.NumberOfVertices; i++)
            {

                if (mGraph.GetWeight(i) > 0)
                {
                    mPartitioning[mNumberOfSubGraphs] = new RSubGraphSD();
                    mPartitioning[mNumberOfSubGraphs].SetGraph(mGraph);
                    mPartitioning[mNumberOfSubGraphs].SetSupplyNode(i);
                    mPartitioning[mNumberOfSubGraphs].InitializeSupplyId();
                    mAvailableSupply += mGraph.GetWeight(i);
                    mSubgraphNodes[i].SubGraphId = mNumberOfSubGraphs;

                    mNumberOfSubGraphs++;
                    if (mNumberOfSubGraphs == mGraph.NumberSupply)
                        break;

                }
            }
            rMaxInt = int.MaxValue;
            RemoveAllSupplyNeighbors();
        }

        public void SetOriginalGraph(RGraphSD nGraph)
        {
            mOriginalGraph = nGraph;
        }
        public void SetGraph(RGraphSD nGraph)
        {
            mGraph = nGraph;
            mSubgraphNodes = new RSubGraphNodeArray(mGraph);
                
                //RRNodeSubgraph[mGraph.NumberOfVertices];

       
        }


        public RSubGraphSD GetSubgraphWithMinSupply()
        {
            List<int> L = GetSubgraphWithMinSupplyList();


            if (L.Count() == 0)
                return null;
            else return mPartitioning[L[0]];
        }





        List<int> GetSubgraphWithMinSupplyList()
        {
            int Min = int.MaxValue;
            List<int> Result = new List<int>();


            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                if ((Min > mPartitioning[i].AvailableSupply) && (mPartitioning[i].AvailableSupply>0))
                {

                    Result.Clear();
                    Result.Add(i);
                    Min = mPartitioning[i].AvailableSupply;

                }
                else
                {
                    if (Min == mPartitioning[i].AvailableSupply)
                    {
                        Result.Add(i);
                    }

                }

            }
            return Result;
        }



        public RSubGraphSD GetSubgraphWithMaxSupply()
        {
            List<int> L = GetSubgraphWithMaxSupplyList();


            if (L.Count() == 0)
                return null;
            else return mPartitioning[L[0]];
        }


        


        List<int> GetSubgraphWithMaxSupplyList()
        {
            int Max = -1;
            List<int> Result = new List<int>();


            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                if (Max < mPartitioning[i].AvailableSupply)
                {

                    Result.Clear();
                    Result.Add(i);
                    Max = mPartitioning[i].AvailableSupply;

                }
                else
                {
                    if (Max == mPartitioning[i].AvailableSupply)
                    {
                        Result.Add(i);
                    }

                }

            }
            return Result;
        }


        public RSubGraphSD GetSubgraphWithMinNieghbors( )
        {
            List<int> L = GetSubgraphWithMinNieghborsList();


            if (L.Count() == 0)
                return null;
            else return mPartitioning[L[0]];
        }
        
        List<int> GetSubgraphWithMinNieghborsList()
        {
            int Min = int.MaxValue;
            List<int> Result = new List<int>();
            int tNN;
            
            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {


                tNN = mPartitioning[i].GetNeighbors().Count();


                if ((Min > tNN) && (tNN > 0))
                {

                    Result.Clear();
                    Result.Add(i);
                    Min = tNN;

                }
                else
                {
                    if (Min == tNN)
                    {
                        Result.Add(i);
                    }

                }

            }
            return Result;
        }


       

        public RSubGraphSD GetSubgraphWithMaxBalanceNeighSup()
        {
            List<int> L = GetSubgraphWithMaxBalanceNeighSupList();


            if (L.Count() == 0)
                return null;
            else return mPartitioning[L[0]];
        }

        List<int> GetSubgraphWithMaxBalanceNeighSupList()
        {
            float Max = -1;
            List<int> Result = new List<int>();
            float tNN;

            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {


                tNN = mPartitioning[i].GetNeighbors().Count();

                if (tNN != 0)
                    tNN = mPartitioning[i].AvailableSupply / tNN;
                else
                    tNN = -1000;

                    if ((Max < tNN) && (tNN > 0))
                    {

                        Result.Clear();
                        Result.Add(i);
                        Max = tNN;

                    }
                    else
                    {
                        if (Max == tNN)
                        {
                            Result.Add(i);
                        }

                    }

            }
            return Result;
        }



        public RSubGraphSD GetSubGraph(int index)
        {
            if (index != -1)
                return mPartitioning[index];
            else
                return null;
        }

        public void Save(System.IO.StreamWriter File)
        {

            String tString;
            List<int> Temp;

            File.WriteLine("Number Of Partitions");
            File.WriteLine(Convert.ToString(mNumberOfSubGraphs));

            for (int i = 0; i < mNumberOfSubGraphs; i++)
            {

                tString = "SupplyNode " + mPartitioning[i].SupplyNodeId;
                File.WriteLine(tString);

                Temp = mPartitioning[i].GetInside();

                foreach (int elem in Temp)
                {
                    tString = elem + " ";
                    File.Write(tString);
                }
                File.WriteLine("");
            }

        }

        public void Save(string FileName)
        {

            System.IO.StreamWriter File = new System.IO.StreamWriter(FileName);
            Save(File);
            File.Close();

        }
    }
}
