﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GraphSD
{
    public class RSubGraphSD
    {

        RGraphSD     mGraph;
        int          mAvailableSupply;
        int          mSupplyNodeId;
        bool         mChanged;
        int          mCoveredDemand;
        List<int>    mInside;
        List<int>    mNeighbors; // Demand;


        public void Changed() { mChanged = true; }

       public  List<int> GetConnectedNodes(int Id, RGraphSD iGraph) {

            List<int> Result = new List<int>();

            foreach(int el in mInside){
            
                    if(iGraph.EdgeExists(el, Id))
                        Result.Add(el);
            }

            return Result;
        
        }

        public int GetCoveredDemand() {

            if (!mChanged)
                return mCoveredDemand;

            int Result = -mGraph.GetWeight(mSupplyNodeId);

            foreach (int elem in mInside)
                Result += mGraph.GetWeight(elem);

            mChanged = false;
            mCoveredDemand = Result;
           
            return Result;
        }
        
        public void RemoveNeighbor(int Index) {

            mNeighbors.Remove(Index);
        }

        public List<int> GetInside() {
            return mInside;
        
        }
        public List<int> GetNeighbors()
        {
            return mNeighbors;
        }
        
        public int SupplyNodeId {
            get { return mSupplyNodeId; }
            set { mSupplyNodeId = value; }
        }

        public void SetGraph(RGraphSD nGraph) { mGraph = nGraph; }

        public void SetSupplyNode(int SupplyId) {
            mSupplyNodeId = SupplyId;
            mAvailableSupply = 0;// mGraph.GetSupply(SupplyId);
        }
        public void InitializeSupplyId(bool removeHighDemand = true) {

            mInside = new List<int>();
            mNeighbors = new List<int>();
            AddDemand(mSupplyNodeId, removeHighDemand);
            
        }
        public int AvailableSupply {

            get { return mAvailableSupply; }
            set { mAvailableSupply = value; }
        }

        public void RemoveHighDemand(){
        
               mNeighbors.RemoveAll(item => mGraph.GetWeight(item) < -AvailableSupply);
        }

        public void AddDemand(int index, bool removeHighDemand=true) {


            mChanged = true;
            mNeighbors.Remove(index);
            if(!mInside.Contains(index))
                            mInside.Add(index);
            mAvailableSupply += mGraph.GetWeight(index);
            
            List<int> Connections = mGraph.GetConnections(index);

            foreach (int con in Connections) {
                if (!mNeighbors.Contains(con))
                                     mNeighbors.Add(con);
            }
            if(removeHighDemand)
                RemoveHighDemand();
        
        }


        public int GetHighestDemand()
        {
            List<int> L = GetHighestDemandList();


            if (L.Count() == 0)
                return -1;
            else return L[0];

        
        }
        public List<int> GetHighestDemandList() {

            int Min = 1;
            List<int> Result = new List<int>();

          
            foreach(int neighbor in mNeighbors){
            
                if((Min>mGraph.GetWeight(neighbor))&& (mGraph.GetWeight(neighbor)<0)){
                    
                    Min = mGraph.GetWeight(neighbor);
                    Result.Clear();
                    Result.Add(neighbor);
                }

                if(Min==mGraph.GetWeight(neighbor))
                      Result.Add(neighbor);

            }
            
            return Result;
        }


        public int GetLowestDemand()
        {
            List<int> L = GetLowestDemandList();


            if (L.Count() == 0)
                return -1;
            else return L[0];


        }
        public List<int> GetLowestDemandList()
        {

            int Max = int.MinValue;
            List<int> Result = new List<int>();


            foreach (int neighbor in mNeighbors)
            {

                if ((Max < mGraph.GetWeight(neighbor)) && (mGraph.GetWeight(neighbor)<0))
                {

                    Max = mGraph.GetWeight(neighbor);
                    Result.Clear();
                    Result.Add(neighbor);
                }

                if (Max == mGraph.GetWeight(neighbor))
                    Result.Add(neighbor);

            }

            return Result;
        }



        //
        //          Corently doesn't fix neighbors and the current graph
        //
        public void RemoveNode(int NodeId) {
            mChanged = true;
            mInside.Remove(NodeId);
        }


        public int GetMaxNeighbors(RSubGraphNodeArray iSubGraphNodes)
        {
            List<int> L = GetMaxNeighborsList(iSubGraphNodes);


            if (L.Count() == 0)
                return -1;
            else return L[0];
        }

    List<int> GetMaxNeighborsList(RSubGraphNodeArray iSubGraphNodes)
    {
         
            List<int>    Result = new List<int>();
            int          MaxIndex = -1;
            int          MaxNeighbors=-1;

            int          Neighbors;
            List<int>   Connections;

            foreach (int elem in mNeighbors) {

                if ((mGraph.GetWeight(elem) <= 0))
                {
                    Connections = mGraph.GetConnections(elem);
                    Neighbors = 0;

                    foreach (int con in Connections)
                    {
                        int l = mGraph.GetWeight(mSupplyNodeId) + GetCoveredDemand() + mGraph.GetWeight(con);
                        if (iSubGraphNodes.IsNonLocated(con) && (mGraph.GetWeight(con) <= 0) &&(l>=0))
                            Neighbors++;
                    }

                    if (MaxNeighbors < Neighbors)
                    {

                        MaxNeighbors = Neighbors;
                        Result.Clear();
                        Result.Add(elem);
                    }

                    if (MaxNeighbors == Neighbors)
                        Result.Add(elem);
                }
            
            }

            return    Result;
         
         }


         public int GetMaxBalanceDemNeigh(RSubGraphNodeArray iSubGraphNodes)
        {
            List<int> L = GetMaxBalanceDemNeighList(iSubGraphNodes);


            if (L.Count() == 0)
                return -1;
            else return L[0];
        }



        List<int> GetMaxBalanceDemNeighList(RSubGraphNodeArray iSubGraphNodes)
        {

            List<int> Result = new List<int>();
            int MaxIndex = -1;
            int MaxDemNeighbors = -1;

            int DemNeighbors;
            int Neighbors;

            List<int> Connections;

            foreach (int elem in mNeighbors)
            {

                if ((mGraph.GetWeight(elem) <= 0))
                {
                    Connections = mGraph.GetConnections(elem);
                    Neighbors = 0;

                    foreach (int con in Connections)
                    {
                        int l = mGraph.GetWeight(mSupplyNodeId) +GetCoveredDemand()  + mGraph.GetWeight(con);
                        if (iSubGraphNodes.IsNonLocated(con) && (mGraph.GetWeight(con) <= 0) && (l>= 0))
                            Neighbors++;
                    }
                    DemNeighbors = Math.Abs((Neighbors + 1) * mGraph.GetWeight(elem));
                    if (MaxDemNeighbors < DemNeighbors)
                    {

                        MaxDemNeighbors = DemNeighbors;
                        Result.Clear();
                        Result.Add(elem);
                    }

                    if (MaxDemNeighbors == DemNeighbors)
                        Result.Add(elem);
                }

            }
            return Result;
            
         
         }

}
}